module Algorithms.Complexity.RuntimeComplexity where

import qualified Data.Set as S
import qualified Data.Map as M
import qualified Data.IntSet as IS
import Data.Array
import List (inits, findIndex, maximumBy)

import Control.Parallel.Strategies
import Control.Exception

import Data.Term
import Data.Term.TRSInfo
import Data.Term.Parser
import Convenience
import Data.Term.Rewrite
import Data.Term.UsableRules
import Data.Types.Types


-- distributes n items in m slots
distribute :: Int -> Int -> [[Int]]
distribute 0 0 = return []
distribute _ 0 = []
distribute n 1 = return [n]
distribute n m = do
    first <- [0..n]
    rest <- distribute (n-first) (m-1)
    return $ first:rest

constructorTerms :: (Ord sym, Ord var2) => TRS sym var1 -> TRSInfo sym var1 -> Bool -> MatchInfo sym var2 -> var2 -> [[MatchedTerm sym var2]]
constructorTerms trs ti welltyped mi v = l
  where
    lhsSyms = S.unions $ map  (\(Rule l _) -> getSyms l) trs
    constructors = S.toList $ S.intersection lhsSyms (tiConstructor ti)
    l = map tol [0..]
    types = tiSimpleTypes ti
    tol 0 = [MTVar v]
    tol n = do
        sym <- constructors
        let Just ar = M.lookup sym (tiArity ti)
        sizes <- distribute (n-1) ar
        let SimpleType argtypes outtype = case M.lookup sym types of Just t -> t
        args <- mapM (\(i, t) -> filter (checkType t) (l!!i) >>= return) $ zip sizes argtypes
        return $ mkMTFunApp mi sym args
    checkType t (MTVar _) = True
    checkType t (MTFunApp symid _ _ _) | welltyped =  case M.lookup sym types of Just (SimpleType _ ot) -> t == ot
                                       | otherwise =  True
      where sym = (miSyms mi)!symid

-- basicTerms :: (Ord sym, Ord var2) => TRS sym var1 -> MatchInfo sym var2 -> var2 -> [[MatchedTerm sym var2]]
basicTerms trs ti welltyped mi v = map tol [1..]
  where
    defined = S.toList $ tiDefined ti
    l = constructorTerms trs ti welltyped mi v
    types = tiSimpleTypes ti
    tol 0 = [MTVar v]
    tol n = do
        sym <- defined
        let Just ar = M.lookup sym (tiArity ti)
        sizes <- distribute (n-1) ar
        let SimpleType argtypes outtype = case M.lookup sym types of Just t -> t
        args <- mapM (\(i, t) -> filter (checkType t) (l!!i) >>= return) $ zip sizes argtypes
        return $ mkMTFunApp mi sym args
    checkType t (MTVar _) = True
    checkType t (MTFunApp symid _ _ _) | welltyped =  case M.lookup sym types of Just (SimpleType _ ot) -> t == ot
                                       | otherwise =  True
      where sym = (miSyms mi)!symid

data Complexity
    = Poly Int
    | Unknown
    deriving Eq

instance Show Complexity where
    show Unknown = "?"
    show (Poly 0) = "O(1)"
    show (Poly n) = "O(n^" ++ show n ++ ")"

makeMonotonic :: (Ord a, Bounded a) => [a] -> [a]
makeMonotonic = mM minBound
  where
    mM _ [] = []
    mM m (x:xs) | x < m     = m:(mM m xs)
                | otherwise = x:(mM x xs)

differentiate :: (Num a) => [a] -> [a]
differentiate [] = []
differentiate [x] = []
differentiate (x:x':xs) = x'-x:(differentiate (x':xs))

isConstant :: (Eq a) => [a] -> Bool
isConstant xs = all (t ==) ts
  where (t:ts) = take 4 $ reverse xs

guessComplexity xs = complexity
  where diffs = iterate (makeMonotonic.differentiate) $ makeMonotonic xs
        lists = takeWhile ((>= 5) . length) diffs
        complexity = case findIndex isConstant lists of 
                            Nothing -> Unknown
                            Just i -> Poly i

-- If idxs is an ascending list, gets all elements from ls
-- at positions in idxs.
getIdxs :: [Int] -> [a] -> [a]
getIdxs idxs ls = gi idxs ls 0
  where gi [] _ _ = []
        gi i@(x:xs) (y:ys) o | x==o = y : (gi xs ys (o+1))
                             | otherwise = gi i ys (o+1)

-- buildBasicTermTRS :: (Ord sym) => TRS sym String -> TRS sym String
buildBasicTermTRS trs = basicTrs
  where
        basicTrs = S.toList $ S.union (S.fromList ur) (S.fromList basicLhsRules)
        ur = getIdxs (IS.toAscList urIdxs) trs
        urIdxs = usable2 (transitiveURClosure2 qdp) (S.empty)
        qdp = let q = map (\(Rule l _) -> l) trs in
            buildQdp basicLhsRules q trs True
        info = buildTRSInfo trs
        basicLhsRules = filter (\(Rule l _) -> isBasicLhs l) trs
        isBasicLhs (FunApp _ xs) = all isConstructorTerm xs
        isConstructorTerm t = S.isSubsetOf (getSyms t) (tiConstructor info)

data Info sym = Info {
    termSize :: Int,
    guessedComplexity :: Complexity,
    maximalLength :: Int,
    numberOfBasicTerms :: Int,
    maximalTerm :: Term sym String
} deriving (Show, Eq)

-- iRCguess :: Ord sym => TRS sym String -> [Info sym]
iRCguess trs ti = map buildInfo $ zip btCounts $ inits $ parBuffer 8 rwhnf $ mapTrail0 len bt
  where
    buildInfo (btCount, termsSteps) = Info {
        termSize = length steps,
        guessedComplexity = guessComplexity steps,
        maximalLength = n,
        numberOfBasicTerms = btCount,
        maximalTerm = toTerm mi big
    }
      where
        complexity = guessComplexity steps
        (big, n) = if terms /= [] then last terms else (MTVar "", 0)
        (terms, steps) = unzip termsSteps
    mi = buildMatchInfo trs
    bt = takeWhile ((/=) []) $ basicTerms trs ti True mi ""
    btCounts = 1:map length bt
    len terms = maxSnd (((MTVar "", 0), 0):map (longestDerivation mi) terms)
    longestDerivation mi term = ((term, l), l)
      where
        l = snd $ maxSnd $ lmImRewrite mi term
    maxSnd = maximumBy (\a b -> compare (snd a) (snd b))
    mapTrail0 f = foldr ((:) . f) [((MTVar "", 0), 0)]


terminates trs = seq (maxLength $ iRCguess trs (buildTRSInfo trs)) True
  where
    maxLength = maximumBy (\a b -> compare (maximalLength a) (maximalLength b))
