{-# LANGUAGE FlexibleContexts, GeneralizedNewtypeDeriving #-}

module Data.Term.Rewrite
    ( MatchInfo(miSymMap, miSyms, miFunApps, miRules )
    , MatchedTerm(MTVar, MTFunApp)
    , buildMatchInfo
    , toMatchedTerm
    , mkMTFunApp
    , toTerm
    , rewriteStep
    , lmImRewrite
    )
where

import Data.Map (Map)
import qualified Data.Map as M
import qualified Data.Set as S
import Data.Array
import Data.List (sort)

import qualified Control.Monad.State as SM
-- import qualified Data.ByteString as BS
import Data.Term
import Data.Maybe (mapMaybe)
import Utils
-- import Data.Term.Parser


import Data.IdentShow


newtype SubtermID = SubtermID Int deriving (Show, Eq, Ord, Enum, Ix)
firstSubtermID :: SubtermID
firstSubtermID = SubtermID 0
newtype SymID = SymID Int deriving (Show, Eq, Ord, Enum, IdentShow, Ix)
firstSymID :: SymID
firstSymID = SymID 0
newtype VarID = VarID Int deriving (Show, Eq, Ord, Enum, IdentShow)
firstVarID :: VarID
firstVarID = VarID 0

data MatchInfo sym var = MatchInfo {
    miSymMap :: Map sym SymID,
    miSyms :: Array SymID sym,
    miFunApps :: Array SymID [(SubtermID, [Either SubtermID VarID], Bool)],
    miRules :: Array SubtermID [(Rule sym var, Term SymID VarID)]
} deriving Show

data MatchedTerm sym var
    = MTVar !var
    | MTFunApp {
        mtSym :: !SymID,
        mtSubTerms :: ![MatchedTerm sym var],
        mtSubRedex :: Bool,
        mtMatches :: [(SubtermID, MSubst sym var)]
    }
    deriving (Show, Eq, Ord)

type MSubst sym var = Map VarID (MatchedTerm sym var)

type MContext sym var = MatchedTerm sym var -> MatchedTerm sym var

buildMatchInfo :: (Ord sym, Ord var) => [Rule sym var] -> MatchInfo sym var
buildMatchInfo trs
    = MatchInfo {
        miSymMap = symMap,
        miSyms = array (firstSymID, lastSymID) $ sort $ map (\(a,b) -> (b,a)) $ M.toList symMap,
        miFunApps = array (firstSymID, lastSymID) [ (i, case M.lookup i funAppMap of Nothing -> []; Just x -> x) | i <- [firstSymID..lastSymID] ],
        miRules = array (firstSubtermID, lastSubtermID) [ (i, case M.lookup i ruleMap of Nothing -> []; Just x -> x) | i <- [firstSubtermID..lastSubtermID] ]
    }
  where
    funAppMap = buildMultiMap $ map (\((symid, subids, introsNonLin), subtermid) -> (symid, (subtermid, subids, introsNonLin))) $ M.toList revFunAppMap
    (renamedRulesWithOrig, (symMap, symIDCount)) = SM.runState (renameRules trs) (M.empty, firstSymID)
    lastSymID = pred symIDCount
    renameRules = mapM renameRule
    renameRule orig@(Rule lhs rhs) = do
        lhs' <- renameTerm lhs
        rhs' <- renameTerm rhs
        return (Rule lhs' rhs', orig)
    renameTerm (Var v) = return $ (Var v)
    renameTerm (FunApp sym subs) = do
        symid <- getId sym
        subs' <- mapM renameTerm subs
        return $ FunApp symid subs'
    normalizedRules = map normalizeRule' renamedRulesWithOrig
    normalizeRule' (rule, origRule) = (normalizeRule [firstVarID..] rule, origRule)
    (idRuleList, (revFunAppMap, subtermIDCount)) = SM.runState foldRules (M.empty, firstSubtermID)
    lastSubtermID = pred subtermIDCount
    foldRules = mapM foldRule normalizedRules
    foldRule (Rule lhs rhs, origRule) = do
        Left id <- foldTerm lhs -- only works for lhs /= var
        return (id, (origRule, rhs))
    foldTerm (Var i) = return $ Right i
    foldTerm (FunApp symid subs) = do
        subids <- mapM foldTerm subs
        let introsNonLin = containsDuplicates $ concatMap (S.toList . getVars) subs
        subtermid <- getId (symid, subids, introsNonLin)
        return $ Left subtermid
    ruleMap = buildMultiMap idRuleList


toTerm :: MatchInfo sym var -> MatchedTerm sym var -> Term sym var
toTerm _ (MTVar var) = Var var
toTerm mi (MTFunApp symid subs _ _) = FunApp sym $ map (toTerm mi) subs
  where
    sym = (miSyms mi)!symid


combineMSubsts :: (Eq sym, Ord var) => [MSubst sym var] -> Maybe (MSubst sym var)
combineMSubsts = foldl addSubst (Just M.empty)
  where
    addSubst Nothing t = Nothing
    addSubst (Just s) t
        = if compatible then Just $ M.union s t else Nothing
      where
        compatible = all checkEntry $ M.toList t
        checkEntry (k, v) = case M.lookup k s of Nothing -> True; Just x -> x == v

mkMTFunApp :: (Ord sym, Ord var) => MatchInfo sym var -> sym -> [MatchedTerm sym var] -> MatchedTerm sym var
mkMTFunApp mi sym =  mkMTFunAppRaw mi symid
    where symid = case M.lookup sym (miSymMap mi) of Just id -> id

mkMTFunAppRaw :: (Ord sym, Ord var) => MatchInfo sym var -> SymID -> [MatchedTerm sym var] -> MatchedTerm sym var
mkMTFunAppRaw mi symid subs = MTFunApp symid subs subRedex matches
  where
    matches = mapMaybe tryMarker conditionalMarkers
    conditionalMarkers = (miFunApps mi)!symid
    tryMarker (i, subMarkers, introsNonLin) = do
        subsubsts <- mapM (uncurry getSubstitution) $ zip subs subMarkers
        subst <- if introsNonLin then combineMSubsts subsubsts else return $ M.unions subsubsts
        return (i, subst)
    getSubstitution (MTVar _) (Left _) = Nothing
    getSubstitution t@(MTFunApp _ _ _ matches) (Left i) = lookup i matches
    getSubstitution mterm (Right var) = return $ M.singleton var mterm
    lookup i [] = Nothing
    lookup i ((j, subst):rest) | i == j = Just subst
                               | otherwise = lookup i rest
    matchesLhs = any (\x -> (miRules mi)!x /= []) $ map fst matches
    subRedex = matchesLhs || any isReducable subs
    isReducable (MTVar _) = False
    isReducable (MTFunApp _ _ r _) = r


-- toMatchedTerm :: (Ord sym, Ord var) => MatchInfo sym var -> Term sym var -> MatchedTerm sym var
toMatchedTerm mi (Var var) = MTVar var
toMatchedTerm mi (FunApp sym subs) = mkMTFunApp mi sym $ map (toMatchedTerm mi) subs


subTerms
  :: (Ord sym, Ord var)
    => MatchInfo sym var
    -> Path
    -> MContext sym var
    -> MatchedTerm sym var
    -> [(Path, MContext sym var, MatchedTerm sym var)]
subTerms _ _ _ (MTVar _) = []
subTerms mi path context (MTFunApp sym subterms _ _) = map build $ split3 subterms
  where
    build (pre, term, post) = ([length pre] ++ path, \x -> context $ mkMTFunAppRaw mi sym (rappend pre (x:post)), term)
    rappend [] ys = ys
    rappend (x:xs) ys = rappend xs (x:ys)
    split3 [] = []
    split3 (x:xs) = f [] x xs
    f ys x [] = [(ys, x, [])]
    f ys x (z:zs) = (ys, x, z:zs):(f (x:ys) z zs)




-- substitute :: (Ord sym, Ord var) => MatchInfo sym var -> MSubst sym var -> Term Int Int -> MatchedTerm sym var
substitute mi subst = substitute'
  where
    substitute' t@(Var v) = case M.lookup v subst of Just t -> t
    substitute' (FunApp sym subterms) = mkMTFunAppRaw mi sym $ map substitute' subterms

{-
rewriteStep
  :: (Ord sym, Ord var) =>
    MatchInfo sym var
    -> Strategy
    -> Bool
    -> MatchedTerm sym var
    -> [(Path, Rule sym var, MatchedTerm sym var)]
    -}
rewriteStep mi strategy leftmost = step [] id
  where
    rules i = (miRules mi)!i
    uncurry3 f (a, b, c) = f a b c
    step _ _ (MTVar _) = []
    step _ _ (MTFunApp _ _ False _) = []
    step path context t@(MTFunApp sym subs True matches) = finalRewrites
      where
        rewrite (i, subst) = map (\(rule, rhs) -> (path, rule, context $ substitute mi subst rhs)) $ rules i
        subRewrites = filter (/= []) $ map (uncurry3 step) $ subTerms mi path context t -- the place for CS
        localRewrites = concatMap rewrite matches
        strategyRewrites = case strategy of
            Full -> localRewrites:subRewrites
            Innermost -> case subRewrites of [] -> [localRewrites]; rews -> rews
            Outermost -> case localRewrites of [] -> subRewrites; rews -> [rews]
        finalRewrites = if leftmost then case strategyRewrites of [] -> []; (x:_) -> x else concat strategyRewrites


-- Rewrites terms to normal forms using leftmost innermost strategy
{-
lmImRewrite
  :: (Ord sym, Ord var) =>
    MatchInfo sym var
    -> MatchedTerm sym var
    -> [(MatchedTerm sym var, Int)]
    -}
lmImRewrite mi = r 0
  where
    rules i = (miRules mi)!i
    rewrite n (i, subst) = map (\(rule, rhs) -> (substitute mi subst rhs, n + 1)) $ rules i
    r n t@(MTVar _) = [(t, n)]
    r n t@(MTFunApp _ _ False _) = [(t, n)]
    r n t@(MTFunApp symid subs True matches) = makeUnique $ concatMap check $ do
        subresults <- mapM (r 0) subs
        let (newsubs, steps) = unzip subresults
        -- newsubs must be normal forms now
        let n' = n + sum steps
        let term@(MTFunApp _ _ _ matches) = mkMTFunAppRaw mi symid newsubs
        if mtSubRedex term
            then concatMap (rewrite n') matches
            else return (term, n')
      where
        check res@(MTVar _, _) = [res]
        check res@(MTFunApp _ _ False _, _) = [res]
        check (t@(MTFunApp _ _ True _), n') = r n' t
        narrowTree = True
        makeUnique | narrowTree     = M.toList . M.fromListWith max
                   | otherwise      = id


