{-# LANGUAGE FlexibleContexts #-}
module Data.Types.Types where

import qualified Control.Monad.State as SM
import Data.Map (Map)
import qualified Data.Map as M
import Data.List
import Data.Maybe

import Data.Term
import Data.IdentShow
import Algorithms.Unify
import Utils

import Debug.Trace

d :: Show a => a -> a
d a = trace (show a) a

namesFrom :: String -> [String]
namesFrom l = [ a ++ b | a <- []:(namesFrom l), b <- map (\x -> [x]) l ]

typenames :: [String]
typenames = map (\x -> 't':x) $ namesFrom ['a'..'z']

data Type a =
      Apply (Type a) (Type a)
    | Typevar a
    deriving (Eq, Ord)


instance Unifiable Type where
    matchVar (Typevar x) = Just x
    matchVar _ = Nothing

    eqConsts (Apply _ _) (Apply _ _) = True
    eqConsts _ _ = False

    getSubs (Apply l r) = [l,r]
    getSubs _ = []

    mapVars f (Apply l r) = Apply (mapVars f l) (mapVars f r)
    mapVars f (Typevar x) = Typevar (f x)

    apply f (Apply l r) = Apply (apply f l) (apply f r)
    apply f v@(Typevar x) = case f x of
        Just x' -> x'
        Nothing -> v

instance IdentShow a => Show (Type a) where
    show (Apply l@(Typevar _) r) = show l ++ " -> " ++ show r
    show (Apply l r) = "(" ++ show l ++ ") -> " ++ show r
    show (Typevar f) = identShow f

data Key a b =
      KConstructor a
    | KVariable b
    deriving (Eq, Ord, Show)

type Assumption a b c = Map (Key a b) (Type c)

modifyFst :: (SM.MonadState (t, t1) m) => (t -> t) -> m ()
modifyFst f = SM.modify (\(a, b) -> (f a, b))

modifySnd :: (SM.MonadState (t1, t) m) => (t -> t) -> m ()
modifySnd f = SM.modify (\(a, b) -> (a, f b))

getVar :: (SM.MonadState (a, [b]) t) => t b
getVar = do
    (m, fresh:rest) <- SM.get
    SM.put (m, rest)
    return fresh

getType :: (Ord k, (SM.MonadState (Map k (Type a), [a]) t)) => k -> t (Type a)
getType v = do
    (m, _) <- SM.get
    case M.lookup v m of
        Nothing -> do
            v' <- getVar
            return $ Typevar v'
        Just t -> return t

-- getType :: (SM.MonadState (Assumption a b c, e) t) => t 

assumptionsApply :: Ord a => Map a (Type a) -> Map k (Type a) -> Map k (Type a)
assumptionsApply m a = M.map (applyMap m) a

contains :: Eq a => a -> Type a -> Bool
contains x (Typevar v)
    | x == v        = True
    | otherwise     = False
contains x (Apply l r) = contains x l || contains x r


assumeRight :: Show l => Either l r -> r
assumeRight (Left err) = error (show err)
assumeRight (Right x) = x

-- type inference algorithm for TRS
w :: (SM.MonadState (Map (Key a b1) (Type b), [b]) m,
      Ord a,
      Ord b1,
      Show b,
      Ord b,
      IdentShow b1,
      IdentShow a,
      IdentShow b,
      Show b1,
      Show a) =>
     Term a b1 -> m (Map b (Type b), Type b)
w (Var a) = do
    (m, _) <- SM.get
    case M.lookup (KVariable a) m of
        Nothing -> do
            v <- getVar
            let t = Typevar v
            modifyFst (M.insert (KVariable a) t)
            return (M.empty, t)
        Just t ->
            return (M.empty, t)
w (FunApp a terms) = do
    (m, _) <- SM.get
    substAndTypes <- mapM w terms
    let (substs, termTypes) = unzip substAndTypes
    let bigsubst = foldl combineMaps M.empty substs
    beta <- getVar
    let inferredType = foldr Apply (Typevar beta) termTypes
    case M.lookup (KConstructor a) m of
        Nothing -> do
            modifyFst (M.insert (KConstructor a) inferredType)
            return (bigsubst, Typevar beta)
        Just assumedType -> do
            let mgu = case unify assumedType inferredType of {
                Left err -> error (show (m, err, a, terms, assumedType, inferredType));
                Right sigma -> sigma }
            d ("mgu", inferredType, assumedType, mgu) `seq` modifyFst (assumptionsApply mgu)
            return (combineMaps bigsubst mgu, applyMap mgu (Typevar beta))

{-
inferTypes :: (Ord t,
               Ord k,
               Show t,
               Show k,
               Show v,
               Ord v,
               IdentShow v,
               IdentShow t,
               IdentShow k) =>
              [Rule k t] -> [v] -> [(k, Type v)]
-}
inferTypes rules vars = M.mapKeys (\(KConstructor a) -> a) types
    where
        removeVars = modifyFst (M.filterWithKey isCon)
        isCon (KConstructor _) _ = True
        isCon _                _ = False
        checkRule (Rule l r) = do
            (_, ltype) <- w l
            (_, rtype) <- w r
            let mgu = case unify rtype ltype of {
                Left err -> error (show (l, r, err));
                Right sigma -> sigma }
            (f, _) <- SM.get
            (d ((Rule l r), f, ltype, rtype, mgu)) `seq`
                modifyFst (assumptionsApply mgu)
            removeVars
            (f', _) <- SM.get
            d f' `seq` return ()
        types = fst $ SM.execState (mapM_ checkRule rules) (M.empty, vars)

        {-
            where
                cv (Typevar v) = do
                    (m, _) <- SM.get
                    case M.lookup v m of
                        Nothing -> do
                            v' <- getVar
                            modifyFst (M.insert v v')
                            return $ Typevar v'
                        Just v' -> return (Typevar v')
                cv (Apply subs t) = do
                    subs' <- mapM cv subs
                    t' <- cv t
                    return $ Apply subs' t'
-}


data SimpleType var = SimpleType [var] var deriving Show

data TypePos sym var = TPIn sym Int | TPOut sym | TPVar Int var deriving (Eq, Ord)

-- simple monomorphic type inference
inferSimpleTypes rules symbols arities = types
  where
    constraints = concatMap (uncurry ruleConstraints) $ zip [0..] rules
    ruleConstraints ruleNum (Rule lhs rhs) =
        let (c1, t1) = termConstraints ruleNum lhs in
        let (c2, t2) = termConstraints ruleNum rhs in
        (t1, t2):(c1 ++ c2)
    termConstraints ruleNum (Var x) = ([], TPVar ruleNum x)
    termConstraints ruleNum (FunApp f subs) = (concat (conflicts:subConflicts), TPOut f)
      where
        (subConflicts, argTypes) = unzip $ map (termConstraints ruleNum) subs
        conflicts = map (\(i, t) -> (TPIn f i, t)) $ zip [0..] argTypes
    (typePosMap, _) = SM.execState (mapM_ solveConstraint constraints) (M.empty, 0)
    solveConstraint (l, r) = do
        lid <- getId l
        rid <- getId r
        if lid == rid
            then return ()
            else do
                modifyFst $ M.map (\x -> if x == lid then rid else x)
    types = M.fromList $ map buildType symbols
    buildType f = (f, SimpleType argTypes outType)
      where
        ar = case M.lookup f arities of Just x -> x
        argTypes = mapMaybe (flip M.lookup typePosMap) [ TPIn f i | i <- [0..ar] ]
        outType = case M.lookup (TPOut f) typePosMap of Just x -> x


